from httplib2 import *
from httplib2 import DEFAULT_MAX_REDIRECTS
import json
import logging
import time
from performanceplatform.collector.logging_setup import (
    extra_fields_from_exception)

_MAX_RETRIES = 5


def abort_message(method, uri, status, reason):
    return (
        '{} request for {} failed with code {}, '
        'reason {}. Aborting. See '
        'https://developers.google.com'
        '/analytics/devguides/reporting/core/v3/coreErrors'
        ' for more details'.format(
            method,
            uri,
            status,
            reason))


def notify_operator(method, uri, status, reason):
    logging.error(abort_message(method, uri, status, reason))


def parse_reason(response, content):
    """
    Google return a JSON document describing the error. We should parse this
    and extract the error reason. See
    https://developers.google.com/analytics/devguides/reporting/core/v3/coreErrors
    """
    def first_error(data):
        errors = data['error']['errors']
        if len(errors) > 1:
            # we have more than one error. We should capture that?
            logging.info('Received {} errors'.format(len(errors)))
        return errors[0]

    try:
        json_data = json.loads(content)
        return first_error(json_data)['reason']
    except:
        return response.reason


class ResponseAction:

    def __init__(self):
        self.should_retry = False
        self.retry_info = ''


def GABackoff(response, content, method, uri):
    response_action = ResponseAction()

    if response.status not in [400, 401, 403, 503]:
        return response_action

    reason = parse_reason(response, content)

    if response.status in [400, 401]:
        # --------------------------------------------------------------------
        # Do not retry without fixing the problem
        # notify someone with appropriate actionable data
        # --------------------------------------------------------------------
        notify_operator(method, uri, response.status, reason)
        return response_action

    if response.status == 503:
        # --------------------------------------------------------------------
        # Server returned an error. Do not retry this query more than once.
        # --------------------------------------------------------------------
        # Not sure we can action anything as a result of this? Just note it.
        logging.info(abort_message(method, uri, response.status, reason))
        return response_action

    if response.status == 403:
        if reason in [
            "insufficientPermissions",
            "dailyLimitExceeded",
            "usageLimits.userRateLimitExceededUnreg",
        ]:
            # ----------------------------------------------------------------
            # insufficientPermissions indicates that the user does not have
            # sufficient permissions for the entity specified in the query.
            #
            # Do not retry without fixing the problem. You need to get
            # sufficient permissions to perform the operation on the specified
            # entity.
            # ----------------------------------------------------------------
            # dailyLimitExceeded indicates that user has exceeded the daily
            # quota (either per project or per view (profile)).
            #
            # Do not retry without fixing the problem. You have used up your
            # daily quota. See
            # https://developers.google.com/analytics/devguides/reporting/core/v3/limits-quotas
            # ----------------------------------------------------------------
            # usageLimits.userRateLimitExceededUnreg indicates that the
            # application needs to be registered in the Google Developers
            # Console.
            #
            # Do not retry without fixing the problem. You need to register in
            # Developers Console to get the full API quota.
            # See https://console.developers.google.com/
            # ----------------------------------------------------------------
            notify_operator(method, uri, response.status, reason)
            return response_action
        elif reason in ["userRateLimitExceeded", "quotaExceeded"]:
            # ----------------------------------------------------------------
            # userRateLimitExceeded indicates that the user rate limit has
            # been exceeded. The maximum rate limit is 10 qps per IP address.
            # The default value set in Google Developers Console is 1 qps per
            # IP address. You can increase this limit in the Google Developers
            # Console to a maximum of 10 qps.
            #
            # Retry using exponential back-off. You need to slow down the rate
            # at which you are sending the requests
            # ----------------------------------------------------------------
            # quotaExceeded indicates that the 10 concurrent requests per view
            # (profile) in the Core Reporting API has been reached.
            #
            # Retry using exponential back-off. You need to wait for at least
            # one in-progress request for this view (profile) to complete.
            # ----------------------------------------------------------------

            # fall through to retry handling below
            response_action.should_retry = True
        else:
            # Unhandled 403 status
            notify_operator(method, uri, response.status, reason)
            return response_action

    # If we got this far, then we're going to retry the request. Capture more
    # detail about why.
    retry_info = ('{} request for {} failed with code {}, '
                  'reason {}'.format(method,
                                     uri,
                                     response.status,
                                     reason))
    response_action.retry_info = retry_info

    return response_action


class HttpWithBackoff(Http):

    def __init__(self, cache=None, timeout=None,
                 proxy_info=None,
                 ca_certs=None, disable_ssl_certificate_validation=False,
                 backoff_strategy_predicate=GABackoff):
        dscv = disable_ssl_certificate_validation
        super(HttpWithBackoff, self).__init__(
            cache=cache, timeout=timeout,
            proxy_info=proxy_info,
            ca_certs=ca_certs,
            disable_ssl_certificate_validation=dscv)
        if backoff_strategy_predicate:
            self._backoff_strategy_predicate = backoff_strategy_predicate
        else:
            self._backoff_strategy_predicate = GABackoff

    def request(self,
                uri,
                method="GET",
                body=None,
                headers=None,
                redirections=DEFAULT_MAX_REDIRECTS,
                connection_type=None):

        delay = 10

        for n in range(_MAX_RETRIES):
            response, content = super(HttpWithBackoff, self).request(
                uri,
                method,
                body,
                headers,
                redirections,
                connection_type)

            response_action = self._backoff_strategy_predicate(
                response, content, method, uri)

            if response_action.should_retry:
                retry_info = response_action.retry_info
                retry_info += '(Attempt {} of {})'.format(n + 1, _MAX_RETRIES)
                if n + 1 < _MAX_RETRIES:
                    retry_info += ' Retrying in {} seconds...'.format(delay)
                logging.info(retry_info)
                time.sleep(delay)
                delay *= 2
            else:
                return response, content

        # we made _MAX_RETRIES requests but none worked
        logging.error(
            'Max retries exceeded for {}'.format(uri),
            # HttpLibException doesn't exist but this fits with the format
            # for logging other request failures
            extra={
                'exception_class': 'HttpLibException',
                'exception_message': '{}: {}'.format(response.status,
                                                     response.reason)
            }
        )

        return response, content



from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)

from performanceplatform.collector.ga.core \
    import query_documents_for

from performanceplatform.collector.ga.lib.helper import create_client

from performanceplatform.utils.data_pusher import Pusher


def main(credentials, data_set_config, query, options, start_at, end_at):
    client = create_client(credentials)

    documents = query_documents_for(
        client, query, options,
        data_set_config['data-type'],
        start_at, end_at)

    Pusher(data_set_config, options).push(documents)



import json
import logging
from performanceplatform.utils.data_parser import DataParser

from performanceplatform.utils.datetimeutil \
    import period_range


def query_ga(client, config, start_date, end_date):
    logging.info("Querying GA for data in the period: %s - %s"
                 % (str(start_date), str(end_date)))

    # If maxResults is 0, don't include it in the query.
    # Same for a sort == [].

    maxResults = config.get("maxResults", None)
    if maxResults == 0:
        maxResults = None

    sort = config.get("sort", None)
    if sort == []:
        sort = None

    return client.query.get(
        config["id"].replace("ga:", ""),
        start_date,
        end_date,
        config["metrics"],
        config.get("dimensions"),
        config.get("filters"),
        maxResults,
        sort,
        config.get("segment")
    )


def try_number(value):
    """
    Attempt to cast the string `value` to an int, and failing that, a float,
    failing that, raise a ValueError.
    """

    for cast_function in [int, float]:
        try:
            return cast_function(value)
        except ValueError:
            pass

    raise ValueError("Unable to use value as int or float: {0!r}"
                     .format(value))


def convert_durations(metric):
    """
    Convert session duration metrics from seconds to milliseconds.
    """
    if metric[0] == 'avgSessionDuration' and metric[1]:
        new_metric = (metric[0], metric[1] * 1000)
    else:
        new_metric = metric
    return new_metric


def build_document(item):
    metrics = [(key, try_number(value))
               for key, value in item["metrics"].items()]
    metrics = [convert_durations(metric) for metric in metrics]

    return dict(metrics)


def pretty_print(obj):
    return json.dumps(obj, indent=2)


def build_document_set(results):
    return (build_document(item)
            for item in results)


def query_for_range(client, query, range_start, range_end):
    frequency = query.get('frequency', 'weekly')

    for start, end in period_range(range_start, range_end, frequency):
        for record in query_ga(client, query, start, end):
            yield record


def query_documents_for(client, query, options,
                        data_type, start_date, end_date):
    results = query_for_range(client, query, start_date, end_date)

    results = list(results)
    frequency = query.get('frequency', 'weekly')
    special_fields = add_timeSpan(frequency, build_document_set(results))
    return DataParser(results, options, data_type).get_data(
        special_fields
    )


def add_timeSpan(frequency, special_fields):
    frequency_to_timespan_mapping = {
        'daily': 'day',
        'weekly': 'week',
        'monthly': 'month',
    }
    timespan = frequency_to_timespan_mapping[frequency]
    return [dict(item.items() + [('timeSpan', timespan)])
            for item in special_fields]



from performanceplatform.utils.data_parser import DataParser
from performanceplatform.utils.data_pusher import Pusher
from datetime import datetime


class BaseCollector(object):
    def __init__(self, credentials, query, start_at, end_at):
        self.start_at = start_at
        self.end_at = end_at
        self.user = credentials['user']
        self.password = credentials['password']
        self.query_format = 'json'

    @classmethod
    def parse_date_for_query(cls, date):
        return date.strftime("%Ym%md%d")

    @classmethod
    def parse_standard_date_string_to_date(cls, date_string):
        if type(date_string) == datetime:
            return date_string
        return datetime.strptime(date_string, "%Y-%m-%d")

    def collect(self):
        data = []
        for start_at, end_at in self.get_date_range_for_webtrends():
            response = self._make_request(
                start_at,
                end_at)
            response.raise_for_status()
            data.append(response.json()["data"])
        return data

    def collect_parse_and_push(self, data_set_config, options):
        raw_json_data = self.collect()
        parsed_data = self.build_parser(
            data_set_config, options).parse(raw_json_data)
        Pusher(data_set_config, options).push(parsed_data)

    def get_date_range_for_webtrends(self):
        return self.__class__.date_range_for_webtrends(
            self.start_at, self.end_at)


class BaseParser(object):
    def __init__(self, options, data_type):
        self.options = options
        self.data_type = data_type

    def parse(self, data):
        base_items = []
        special_fields = []
        for item in data:
            res = self.parse_item(item)
            base_items += res[0]
            special_fields += res[1]
        return DataParser(
            base_items, self.options, self.data_type
        ).get_data(special_fields)
